import json
import logging
import fsspec

from pathlib import Path
from typing import Dict, List

from io import IOBase
from plum import dispatch


logger = logging.getLogger(__name__)


@dispatch
def read_records(obj: IOBase) -> List[Dict]:
    """Reads a json file from an IOBase object"""
    if not obj.size:
        return []
    if not obj.readable():
        return []
    data = json.loads(obj.read())
    return read_records(data)


@dispatch
def read_records(obj: List[IOBase]) -> List[Dict]:
    """Reads a json file from a ist of IOBase objects"""
    docs = []
    for f in obj:
        docs.extend(read_records(f))
    return docs


@dispatch
def read_records(obj: Path) -> List[Dict]:
    """Reads a json file"""
    data = json.loads(obj.read_text())
    return read_records(data)


@dispatch
def read_records(obj: List[Dict]) -> List[Dict]:
    """This is the expected format for a list of documents"""
    return obj


@dispatch
def read_records(obj: Dict[str,List]) -> List[Dict]:
    docs = []
    for _, data in obj.items():
        docs.extend(data)
    return docs


@dispatch
def read_records(obj: Dict[str,Dict]) -> List[Dict]:
    """Reads a dict of dicts, assume its tinydb format"""
    docs = []
    for data in obj.values():
        for d in data.values():
            docs.append(d)
    return docs


class JsonLoader:
    def __init__(self, path, **storage_kwargs):
        self.path = path
        self.storage_kwargs = storage_kwargs

    def read(self) -> List[dict]:
        with fsspec.open_files(self.path, **self.storage_kwargs) as fs:
            if len(fs):
                docs = read_records(fs)
            else:
                docs =[]
        return docs

    def __call__(self) -> List[dict]:
        return self.read()



"""
# S1 (x,y,z) correction
S1 signals in LXeTPC has a large position dependence across the whole volume. In order to make a precise energy reconstruction, we need a “correction” map to overcome its spatial dependence, as the case for S2.

See [description in the Team C overview page](https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:analysis:reconstruction_team#s1_x_y_z_correction)

The jupyter notebook in this folder is replaced by [this](https://github.com/XENONnT/nton/blob/master/nton/analyses/corrections/s1correction.py) script in the nton file

"""

from ..base_references import BaseMap



class S1XYZMap(BaseMap):
    _ALIAS = "s1_xyz_maps"
    _ALLOWED_FORMATS = ["json.gz", "json"]




"""
# Photoionization strengths
https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt_sr1:photoionization_origin
"""

from ..base_corrections import TimeSampledCorrection


class PhotoionizationStrength(TimeSampledCorrection):
    _ALIAS = "photoionization_strengths"
    value: float



import os
import parse
import fsspec

import pandas as pd

from typing import Any, Union
from collections import UserDict
from pydantic import ValidationError
from rframe.data_accessor import DataAccessor
from typing import List
from plum import dispatch

from .dispatchers import read_files


def docs_to_wiki(schema, docs, title=None, columns=None):
    """Convert a list of documents to a dokuwiki table

    :param title: title of the table.
    """
    if title is None:
        title = schema._ALIAS.replace("_", " ").capitalize() + " Table"

    if columns is None:
        columns = list(schema.get_index_fields()) + list(schema.get_column_fields())

    table = "^ %s " % title + "^" * (len(columns) - 1) + "^\n"
    table += "^ " + " ^ ".join(columns) + " ^\n"

    for doc in docs:
        table += "| " + " | ".join([str(getattr(doc, col)) for col in columns]) + " |\n"
    return table


def docs_to_dataframe(schema, docs, columns=None):
    if columns is None:
        columns = list(schema.__fields__)
    else:
        columns = list(columns)

    docs = [doc.pandas_dict() for doc in docs]
    df = pd.json_normalize(docs)

    if len(df):
        df = df[columns]
    else:
        df = df.reindex(columns=list(columns))

    index_fields = [name for name in schema.get_index_fields() if name in columns]

    if len(index_fields) == 1:
        index_fields = index_fields[0]

    return df.set_index(index_fields)


class LazyDataAccessor(DataAccessor):
    __storage__ = None

    @property
    def storage(self):
        if self.__storage__ is None:
            self.__storage__ = self.get_storage()
        if not isinstance(self.__storage__, list):
            return self.__storage__
        docs = []
        for doc in self.__storage__:
            if not isinstance(doc, dict):
                continue
            try:
                doc = self.schema(**doc).pandas_dict()
                docs.append(doc)
            except ValidationError:
                continue

        df = pd.DataFrame(docs, columns=list(self.schema.__fields__))
        idx_names = list(self.schema.get_index_fields())
        if not all([n in df.columns for n in idx_names]):
            return df
        if len(idx_names) == 1:
            idx_names = idx_names[0]
        df = df.set_index(idx_names)
        self.__storage__ = df
        return df

    @storage.setter
    def storage(self, value):
        if not callable(value):
            raise ValueError('storage must be callable')
        self.get_storage = value


def remove_prefix(text, prefix, sep="/"):
    if text.startswith(prefix):
        text = text[len(prefix):]
    if sep is not None and text.startswith(sep):
        text = text[len(sep):]
    return text


class LazyFileAccessor(DataAccessor):
    loaded: set
    pattern: str
    protocol: str
    root: str
    urlpaths: List[str]
    storage_options: dict
    
    def __init__(self, schema, urlpaths, **kwargs):
        if isinstance(urlpaths, str):
            urlpaths = [urlpaths]
        self.loaded = set()
        self.storage_options = kwargs
        self.urlpaths = urlpaths
        storage = schema.empty_dframe()
        super().__init__(schema, storage, False)

    def _find(self, skip=None, limit=None, sort=None, **labels):
        self.load_files(**labels)
        return super()._find(skip=skip, limit=limit, sort=sort, **labels)

    def format_to_glob(self, path):
        for field in self.schema.__fields__:
            test_cases = ["{" + field + "}", "{ " + field + " }", "{" + field + ":s}", "{" + field + ":d}"]
            for test_case in test_cases:
                if test_case in path:
                    path = path.replace(test_case, "*")
        return path

    def glob_to_format(self, path):
        return path.replace("*", "{}")

    def iter_path_records(self, ignore_paths=(), **labels):
        for path in self.urlpaths:
            glob_patttern = self.format_to_glob(path)
            fs, _, fpaths = fsspec.get_fs_token_paths(glob_patttern, storage_options=self.storage_options)
            # force the protocol to be file
            fs.protocol = "file"
            pattern = path.replace(f"{fs.protocol}://", "")
            pattern = parse.compile(self.glob_to_format(pattern))
            loaded = set(ignore_paths)

            for fpath in fpaths:
                if fpath in loaded:
                    continue
                r = pattern.parse(fpath)
                if r is None:
                    continue
                for k,vs in labels.items():
                    
                    if vs is None:
                        continue
                    if not isinstance(vs, list):
                        vs = [vs]
                    elif not len(vs):
                        continue
                    label = r.named.get(k, None)
                    if label is None:
                        continue
                    if k in self.schema.__fields__:
                        index = self.schema.index_for(k)
                        label = index.validate_label(label)
                        vs = [index.validate_label(v) for v in vs]
                    if label not in vs:
                        break
                else:
                    records = read_files(fpath, protocol=fs.protocol, **fs.storage_options)
                    yield fpath, records
                    self.loaded.add(fpath)

    def load_files(self, **labels):
        index_fields = list(self.schema.get_index_fields())
        if len(index_fields) == 1:
            index_fields = index_fields[0]
        docs = []
        for fpath, records in self.iter_path_records(ignore_paths=self.loaded, **labels):
            for doc in records:
                if not isinstance(doc, dict):
                    continue
                try:
                    doc = self.schema(**doc).pandas_dict()
                    docs.append(doc)
                except ValidationError as e:
                    continue
            self.loaded.add(fpath)
        df = pd.DataFrame(docs, columns=list(self.schema.__fields__))
        df = df.set_index(index_fields)
        self.storage = pd.concat([self.storage, df]).sort_values(index_fields)

    def _min(self, **kwargs):
        self.load_files()
        return super()._min(**kwargs)

    def _max(self, **kwargs):
        self.load_files()
        return super()._max(**kwargs)

    def _count(self, **kwargs):
        self.load_files()
        return super()._count(**kwargs)

    def _unique(self, **kwargs):
        self.load_files()
        return super()._unique(**kwargs)

    def insert(self, docs, raise_on_error=True, dry=False):
        raise NotImplementedError

    def delete(self, docs, raise_on_error=True):
        raise NotImplementedError

class FileLoader:
    root: str = None
    protocol: str = "file"
    loaded: set
    pattern: str
    root: str
    urlpaths: List[str]
    storage_options: dict

    @property
    def prefix(self):
        return f"{self.protocol}://{self.root}/"
    
    def abs_path(self, path):
        if path.startswith(self.prefix):
            return path
        return f"{self.prefix}{path}"

    def rel_path(self, path):
        if path.startswith(self.prefix):
            return path[len(self.prefix):]
        return path

    def load_files(self, **labels):
        dfs = []
        index_fields = list(self.schema.get_index_fields())
        if len(index_fields) == 1:
            index_fields = index_fields[0]
        for urlpath in self.urlpaths:
            fs, _, paths = fsspec.get_fs_token_paths(urlpath, storage_options=self.storage_options)
            for path in self.filter_paths(paths, **labels):
                if path in self.loaded:
                    continue
                records = read_files(path, **self.storage_options)
                df = pd.json_normalize(records).set_index(index_fields)
                dfs.append(df)
                self.loaded.add(path)
        self.storage = pd.concat([self.storage] + dfs).sort_values(index_fields)



class Database(UserDict):
    def __getattr__(self, attr):
        if attr in self.keys():
            return self[attr]
        raise AttributeError(attr)

    def __dir__(self):
        return list(self.keys())

    def __getitem__(self, key: Any) -> Any:
        if hasattr(key, "_ALIAS"):
            key = key._ALIAS
        return super().__getitem__(key)



from .base_schemas import VersionedXeDoc, XeDoc
from .detector_numbers import DetectorNumber
from .context_config import ContextConfig
from .context_lineage import ContextLineage
from .historian_reader import HistorianTag
from .fax import FaxConfig
from .corrections import *
from .calibrations import *
from .operations_reports import *
from .analysis import *
from .pmt_data import *



"""
Electron Drift Velocity and Drift Time at Gate

The Understanding the shape of S2s is a prerequisite for calibrating the trigger, processor, and the S2 width cut. Drift time and velocity are important parameters needed for the event z-reconstruction and to understand the S2 shape.
The drift time is calculated as the time difference between events originate at gate and cathode. The drift velocity is then extracted dividing the XENONnT cathode-gate distance of 1485 mm. Since z coordinate starts at the gate,
while the drift time starts at the liquid level, therefore we need to subtract the electron drift time at the gate to get
the proper z coordinate, as well as to remove events happening above the gate.

See note on drift velocity, diffusion constant, and drift time at the gate
    https://xe1t-wiki.lngs.infn.it/doku.php?id=dandrea:diffusionstudywithkrdata

"""

import rframe

from ..base_corrections import TimeSampledCorrection


class ElectronDriftVelocity(TimeSampledCorrection):
    _ALIAS = "electron_drift_velocities"
    value: float


class DriftTimeGate(TimeSampledCorrection):
    _ALIAS = "electron_drift_time_gates"
    value: float



"""
S1 Pattern Maps

https://xe1t-wiki.lngs.infn.it/doku.php?id=min:s1opticalmap_comparison (Min + Giovanni)

"""

from ..base_references import BaseMap


class S1PatternMap(BaseMap):
    _ALIAS = "s1_pattern_maps"
    fmt = "pkl"



import fsspec
import rframe
import pandas as pd

from io import IOBase
from typing import Any, List
from plum import dispatch

from xedocs.json_records import read_records

from .regex_dispatcher import RegexDispatcher


@dispatch
def json_serializable(value: Any):
    return value


@dispatch
def json_serializable(value: list):
    return [json_serializable(v) for v in value]


@dispatch
def json_serializable(value: tuple):
    return tuple(json_serializable(v) for v in value)


@dispatch
def json_serializable(value: dict):
    return {json_serializable(k): json_serializable(v) for k, v in value.items()}


@dispatch
def json_serializable(value: pd.Interval):
    return json_serializable((value.left, value.right))


@dispatch
def json_serializable(value: rframe.IntegerInterval):
    return value.left, value.right


@dispatch
def json_serializable(value: rframe.TimeInterval):
    return f"{str(value.left)} to {str(value.right)}"


read_files = RegexDispatcher('read_files')


@read_files.register(r'.*\.csv')
def read_csv_files(path, **kwargs) -> List[dict]:
    docs =[]
    
    with fsspec.open_files(path, **kwargs) as fs:
        kwargs = rframe.utils.filter_kwargs(pd.read_csv, kwargs)
        for f in fs:
            ds = pd.read_csv(f, **kwargs).to_dict(orient='records')
            docs.extend(ds)
    return docs


@read_files.register(r'.*\.json')
def read_json_files(path, **kwargs) -> List[dict]:
    with fsspec.open_files(path, **kwargs) as fs:
        docs = read_records(fs)
    return docs


@read_files.register(r'.*\.parquet')
@read_files.register(r'.*\.pq')
def read_parquet_files(path, **kwargs) -> List[dict]:
    docs =[]
    
    with fsspec.open_files(path, **kwargs) as fs:
        kwargs = rframe.utils.filter_kwargs(pd.read_parquet, kwargs)
        for f in fs:
            ds = pd.read_parquet(f, **kwargs).to_dict(orient='records')
            docs.extend(ds)
    return docs

@read_files.register(r'.*\.xlsx')
@read_files.register(r'.*\.xls')
def read_excel_files(path, **kwargs) -> List[dict]:
    docs =[]
    with fsspec.open_files(path, **kwargs) as fs:
        kwargs = rframe.utils.filter_kwargs(pd.read_excel, kwargs)
        for f in fs:
            ds = pd.read_excel(f, **kwargs).to_dict(orient='records')
            docs.extend(ds)
    return docs


@read_files.register(r'.*\.pkl')
def read_pickle_files(path, **kwargs) -> List[dict]:
    docs =[]
    with fsspec.open_files(path, **kwargs) as fs:
        kwargs = rframe.utils.filter_kwargs(pd.read_pickle, kwargs)
        for f in fs:
            ds = pd.read_pickle(f, **kwargs).to_dict(orient='records')
            docs.extend(ds)
    return docs



"""
# cS2 bottom top ratio correction
Due to photoionizations, the cS2_bottom has a time-dependent charge yield.

https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:zihao:sr1_s2aft_photonionization_correction

"""

from ..base_corrections import TimeSampledCorrection


class cS2BottomTopRatioCorrection(TimeSampledCorrection):
    _ALIAS = "cs2_bottom_top_ratios"
    value: float



"""

"""
import datetime
from typing import Literal
import rframe

from ..base_references import BaseResourceReference


class NaiveBayesClassifier(BaseResourceReference):
    _ALIAS = "bayes_models"
    fmt = "binary"

    value: str



import rframe
import datetime

from .corrections.base_corrections import TimeIntervalCorrection
from .constants import PARTITION

class DetectorNumber(TimeIntervalCorrection):
    """Detector parameters
    A collection of non-time dependent detector
    values.
    """

    _ALIAS = "detector_numbers"

    field: str = rframe.Index(max_length=80)
    partition: PARTITION = rframe.Index(default="all_tpc")
    science_run: str = rframe.Index(max_length=80)
    definition: str=""
    value: float
    uncertainty: float
    reference: str = ""




from typing import Literal
import rframe
from .base_schemas import XeDoc
from pydantic import validator


class HistorianTag(XeDoc):
    """Schema for historian tags to read into influxdb for grafana
    """
    _ALIAS = "historian_tags"
    _CATEGORY = "operations"

    tag_name: str = rframe.Index() # this is the tag name to read from in the historian
    measurement_name: str = rframe.Index() # this is the measurement name to use in influxdb
    bucket_name: str = rframe.Index() # this is the bucket name to use in influxdb
    
    description: str = ""
    
    query_type: Literal["LAB","RAWBYTIME"] = "LAB"
    query_interval: int = 1 # in seconds
    
    category: str = ""
    subsystem: str = ""


    @validator("measurement_name")
    def validate_measurement_name(cls, v, values):
        if v is None:
            v = values["tag_name"]
        return v



import time
import datetime
import numbers
import os
from typing_extensions import dataclass_transform
import unittest
from typing import List

import pandas as pd
import pydantic
import pymongo
import pytz
from hypothesis import assume, given, settings
from hypothesis import strategies as st
from rframe.schema import InsertionError, UpdateError
from xedocs.schemas import (
    BaseCorrectionSchema,
    TimeSampledCorrection,
    TimeIntervalCorrection,
)
import xedocs


def round_datetime(dt):
    return dt.replace(microsecond=int(dt.microsecond / 1000) * 1000, second=0)


# enforce datetimes are in the a reasonable range
# pandas nanosecond resolution has trouble with extreme dates
datetimes = st.datetimes(
    min_value=datetime.datetime(2000, 1, 1, 0, 0),
    max_value=datetime.datetime(2231, 1, 1, 0, 0),
).map(round_datetime)
floats = st.floats(
    min_value=-1e10, max_value=1e10, allow_nan=False, allow_infinity=False
)


def mongo_uri_not_set():
    return "TEST_MONGO_URI" not in os.environ


def api_server_not_set():
    return "TEST_CMT_SERVER_URI" not in os.environ


def make_datetime_index(start, stop, step="1d"):
    if isinstance(start, numbers.Number):
        start = pd.to_datetime(
            start,
            unit="s",
        )
    if isinstance(stop, numbers.Number):
        stop = pd.to_datetime(stop, unit="s")
    return pd.date_range(start, stop, freq=step)


def make_datetime_interval_index(start, stop, step="1d"):
    if isinstance(start, numbers.Number):
        start = pd.to_datetime(start, unit="s")
    if isinstance(stop, numbers.Number):
        stop = pd.to_datetime(stop, unit="s")
    return pd.interval_range(start, stop, periods=step)


@st.composite
def non_overlapping_interval_lists(draw, elements=st.datetimes(), min_size=2):
    elem = draw(st.lists(elements, unique=True, min_size=min_size * 2).map(sorted))
    return list(zip(elem[:-1:2], elem[1::2]))


@st.composite
def non_overlapping_interval_ranges(draw, elements=st.datetimes(), min_size=2):
    elem = draw(st.lists(elements, unique=True, min_size=min_size).map(sorted))
    return list(zip(elem[:-1], elem[1:]))


def pmt_to_pe_space(datetime_range=datetime.timedelta(days=100)):
    utcnow = datetime.datetime.utcnow()
    data = dict(
        version=["ONLINE"] + ["v{i}" for i in range(10)],
        detector=["tpc", "neutron_veto", "muon_veto"],
        pmt=list(range(100)),
        time=make_datetime_index(utcnow - datetime_range, utcnow + datetime_range),
    )


class SimpleCorrection(BaseCorrectionSchema):
    _ALIAS = "simple_correction"

    value: float


class SomeSampledCorrection(TimeSampledCorrection):
    _ALIAS = "sampled_correction"

    value: float


class SomeTimeIntervalCorrection(TimeIntervalCorrection):
    _ALIAS = "time_interval_correction"

    value: float


@st.composite
def time_interval_corrections_strategy(draw, **overrides):
    docs = draw(
        st.lists(
            st.builds(
                SomeTimeIntervalCorrection,
                time=datetimes,
                created_date=datetimes,
                **overrides,
            ),
            min_size=1,
            unique_by=lambda x: (x.version, x.time.left),
        )
    )

    last = docs[-1].time.left + datetime.timedelta(days=10)
    times = sorted([doc.time.left for doc in docs]) + [last]

    for doc, left, right in zip(docs, times[:-1], times[1:]):
        # very small differences may be hard to test for correctness
        assume((right - left) > datetime.timedelta(seconds=10))

        doc.time = left, right

    return docs


class TestCorrections(unittest.TestCase):
    """
    Requires write access to some pymongo server, the URI of witch is to be set
    as an environment variable under:

        TEST_MONGO_URI

    """

    _run_test = True

    def setUp(self):
        # Just to make sure we are running some mongo server, see test-class docstring
        if "TEST_MONGO_URI" not in os.environ:
            self._run_test = False
            return
        uri = os.environ.get("TEST_MONGO_URI")
        db_name = "test_cmt2"
        db = pymongo.MongoClient(uri)[db_name]

        self.collections = {name: db[name] for name in xedocs.list_schemas()}
        for collection in self.collections.values():
            collection.drop()

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    def tearDown(self):
        for collection in self.collections.values():
            collection.drop()

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    @given(
        st.lists(
            st.builds(SomeSampledCorrection, version=st.just("v1"), value=floats),
            min_size=3,
            unique_by=lambda x: x.time,
        )
    )
    @settings(deadline=None)
    def test_sampled_correction_v1(self, docs: List[SomeSampledCorrection]):
        name = SomeSampledCorrection._ALIAS
        datasource = self.collections[name]
        datasource.delete_many({})

        # we must sort by time before inserting

        docs = sorted(docs, key=lambda x: x.time)

        for doc1, doc2 in zip(docs[:-1], docs[1:]):
            # Require minimum 10 second spacing between samples
            # otherwise we get into trouble with rounding
            assume((doc2.time - doc1.time) > datetime.timedelta(seconds=10))

        # since values are interpolated in time
        prev = 0.0
        for doc in docs:
            if abs(doc.value - prev) < 1:
                doc.value += 2
            prev = doc.value
            doc.save(datasource)

        df_original = pd.DataFrame([doc.pandas_dict() for doc in docs]).set_index(
            ["version", "time"]
        )

        for doc1, doc2 in zip(docs[:-1], docs[1:]):
            dt = doc1.time + (doc2.time - doc1.time) / 2

            doc_interp = SomeSampledCorrection.find_one(datasource, time=dt)

            assert (
                doc_interp is not None
            ), f"No interpolated value found for date {dt} in {list(datasource.find())}"

            half_value = (doc2.value + doc1.value) / 2

            # avoid any division by small numbers
            if abs(doc_interp.value) < 1e-9:
                continue

            error = abs(half_value - doc_interp.value) / abs(doc2.value - doc1.value)

            self.assertAlmostEqual(error, 0, delta=1e-2)

            new_doc = SomeSampledCorrection(
                version="v1",
                time=dt,
                value=2 * (half_value + 10),
                created_date=doc_interp.created_date,
            )

            # changing values is not allowed
            with self.assertRaises(UpdateError):
                new_doc.save(datasource)

            # saving the same value as existing document should be allowed
            new_doc.value = doc_interp.value
            new_doc.save(datasource)

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    @given(
        st.lists(
            st.builds(
                SomeSampledCorrection,
                version=st.just("ONLINE"),
                run_id=datetimes,
                value=floats,
            ),
            min_size=3,
            unique_by=lambda x: str(x.time),
        )
    )
    @settings(deadline=None)
    def test_sampled_correction_online(self, docs: List[SomeSampledCorrection]):
        name = SomeSampledCorrection._ALIAS
        datasource = self.collections[name]
        datasource.delete_many({})

        # we must sort by time before inserting
        # since values are interpolated in time
        docs = sorted(docs, key=lambda x: x.time)

        for doc in docs:
            clock = xedocs.settings.clock

            if clock.after_cutoff(doc.time, buffer=2):
                doc.save(datasource)

            # If the time is before the cutoff, should raise an error
            elif not clock.after_cutoff(doc.time):
                current = SomeSampledCorrection.find_one(datasource, **doc.index_labels)
                if current and current.same_values(doc):
                    doc.save(datasource)
                else:
                    error = UpdateError if current else InsertionError
                    with self.assertRaises(error):

                        doc.save(datasource)
                    # insert data manually for testing
                    datasource.insert_one(doc.dict())
                if not clock.after_cutoff(doc.time, buffer=-5):
                    now = clock.current_datetime()
                    found = SomeSampledCorrection.find(datasource, time=now)
                    self.assertLessEqual(1, len(found))

    @unittest.skipIf(mongo_uri_not_set(), "No access to test database")
    @given(time_interval_corrections_strategy(version=st.just("v1"), value=floats))
    @settings(deadline=None)
    def test_interval_correction_v1(self, docs: List[SomeTimeIntervalCorrection]):
        name = SomeTimeIntervalCorrection._ALIAS
        datasource = self.collections[name]
        datasource.delete_many({})

        clock = xedocs.settings.clock

        for doc in docs:
            doc.save(datasource)

            dt = doc.time.left + (doc.time.right - doc.time.left) / 2
            doc_found = SomeTimeIntervalCorrection.find_one(datasource, time=dt)
            self.assertEqual(doc.value, doc_found.value)

        for doc in docs[:-1]:
            dt = doc.time.left + (doc.time.right - doc.time.left) / 2
            doc_found = SomeTimeIntervalCorrection.find_one(datasource, time=dt)
            self.assertEqual(doc.value, doc_found.value)

            doc_found.value += 1
            if clock.after_cutoff(doc.time.left):
                doc_found.save(datasource)
            else:
                with self.assertRaises(UpdateError):
                    doc_found.value = 2 * doc_found.value + 1
                    doc_found.save(datasource)
            doc.save(datasource)

        last_doc = docs[-1]
        left, right = last_doc.time.left, last_doc.time.right

        if not clock.after_cutoff(left) and clock.after_cutoff(right):
            cutoff = clock.cutoff_datetime()
            half_diff = (cutoff - clock.normalize_tz(left)) / 2
            if half_diff > datetime.timedelta(seconds=1):
                with self.assertRaises(UpdateError):
                    last_doc.time = (left, left + half_diff)
                    last_doc.save(datasource)
        else:
            last_doc.time = left, right + datetime.timedelta(days=1)
            last_doc.save(datasource)



from .base_report import BaseOperationsReport


class AbnormalDAQRate(BaseOperationsReport):
    """Abnormal DAQ rate report"""

    _ALIAS = "abnormal_daq_rates"

    anode_voltage_kv: float
    action_taken: str
    plot: str

    reader0_rate: float
    reader1_rate: float
    reader2_rate: float
    muon_veto_rate: float
    nveto_rate: float



"""
Correction: Average single electron gain
Affects: Corrected areas

Average of the single electron corrections. This correction my also have two partitions as this average is a temporal average not a spatial one.

wiki:
"""

import rframe

from ..base_corrections import TimeIntervalCorrection
from ...constants import PARTITION


class AvgSingleElectronGain(TimeIntervalCorrection):

    _ALIAS = "avg_se_gains"
    partition: PARTITION = rframe.Index(default="all_tpc")
    value: float



import datetime
from pydantic import conint
from typing import Literal
from .base_calibrations import BaseCalibration
from ..constants import DIFFUSED_SOURCE


class DiffusedCalibration(BaseCalibration):
    """Internal diffused Calibrations"""

    _ALIAS = "diffused_calibrations"

    source: DIFFUSED_SOURCE

    nv_ticks: conint(ge=0)
    valve_opened: datetime.datetime
    valve_closed: datetime.datetime
